@@ -27,7 +27,8 @@ from typing import Any, List
 import enum
 
 from .base_sliver import BaseSliver
-from .switch_fabric import SwitchFabricInfo
+from .network_service import NetworkServiceInfo
+from .topology_diff import TopologyDiff
 
 
 @enum.unique
@@ -37,6 +38,7 @@ class ComponentType(enum.Enum):
     SharedNIC = enum.auto()
     FPGA = enum.auto()
     NVME = enum.auto()
+    Storage = enum.auto()
 
     def __repr__(self):
         return self.name
@@ -49,10 +51,10 @@ class ComponentSliver(BaseSliver):
 
     def __init__(self):
         super().__init__()
-        self.switch_fabric_info = None
+        self.network_service_info = None
 
-    def set_switch_fabric_info(self, sf_info: SwitchFabricInfo):
-        self.switch_fabric_info = sf_info
+    def set_network_service_info(self, ns_info: NetworkServiceInfo):
+        self.network_service_info = ns_info
 
     @staticmethod
     def type_from_str(ctype: str) -> ComponentType or None:
@@ -62,10 +64,13 @@ class ComponentSliver(BaseSliver):
             if ctype == str(t):
                 return t
 
+    def diff(self, other_sliver) -> TopologyDiff or None:
+        raise RuntimeError('Not implemented')
+
 
 class AttachedComponentsInfo:
     """
-    Stores attached components as a dictionary by PCI ID
+    Stores attached components as a dictionary by PCI ID and name
     """
     def __init__(self):
         self.devices = {}
